# import numpy
import grid
# import scipy
import collections
# import scipy.optimize
import pygdh_io
import time
# import random
import problem

class Coordinator:

    def initialize_grids(self,grid_sequence,past_timestep_count=0):
        """Initialize Coordinator object.

        **Parameter**

        ``grid_sequence``: a ``list`` or ``tuple`` of ``grid.Grid`` objects
        associated with spatial domains.  The index of an object in this
        sequence becomes the integer label associated with the object.


        ``past_timestep_count``: an integer giving the number of past
        solutions that are used in the time discretization formulas
        appearing in equations associated with the present ``Grid``        
        """

        self.past_timestep_count = past_timestep_count
        self.time = collections.deque(range(self.past_timestep_count+1))

        # Check that parameters are tuples or lists
        assert (type(grid_sequence) == type(tuple())
                or type(grid_sequence) == type(list())), '! Coordinator.initialize_problems: expected list or tuple of "Grid" objects as "grid_sequence" parameter.'

        for element in grid_sequence:
            # Check that this is an object of the grid.Grid type
            assert isinstance(element,grid.Grid), '! Coordinator.initialize_problems: expected "Grid" objects as elements of "grid_sequence" parameter'

        # Store argument for later use
        self.grid = grid_sequence
        
        self.grid_count = len(grid_sequence)

        # Each grid.Grid object will be informed of its integer label
        for grid_number in range(self.grid_count):
            self.grid[grid_number].allocate_storage(self.past_timestep_count)
            self.grid[grid_number].number = grid_number

        self.input = None
            
    def initialize_problems(self,problem_sequence):
    
        assert (type(problem_sequence) == type(tuple())
                or type(problem_sequence) == type(list())), '! Coordinator.initialize_problems: expected list or tuple of "Grid" objects as "problem_sequence" parameter.'

        for element in problem_sequence:
            # Check that this is an object of the grid.Grid type
            assert (isinstance(element,problem.Problem)
                    or type(element) == type(tuple())
                or type(element) == type(list())), '! Coordinator.initialize_problems: expected "Problem," tuple, or list objects as elements of "problem_sequence" parameter'

        self.problem = []
        self.frequency = {}
        for element in problem_sequence:
            if type(element) == type(tuple()) or type(element) == type(list()):
                assert (isinstance(element[0],problem.Problem) and type(element[1]) == int), '! Coordinator.initialize_problems: expected (Problem, int) or [Problem, int] as elements of "problem_sequence" parameter'
                self.problem.append(element[0])
                self.frequency[element[0]] = element[1]
            else:
                self.problem.append(element)
                self.frequency[element] = 1
 
        self.time = collections.deque(range(self.past_timestep_count+1))
       
    def solve_time_dependent_system(self,output_filename_root,output_types,maximum_time_interval=None,timestep_size=None,timestep_count=None,handle_exception_in_solver=True,adaptive_timestepping=False,minimum_timestep_size=1e-6,use_predictor=True,quiet=False):
        """Solve time-dependent system.
        
        **Parameters**

        ``restart``: ``False`` if this is a new simulation, ``True`` if restarting from saved data
        
        ``output_filename_root``: a string that will be used at the start of all
        names of files created for output

        ``output_types``: a ``list`` of strings specifying the desired data
        output formats.  Currently available options are ``'HDF5'``, ``'CSV'``,
        and ``'GNUPLOT'``

        **Returns** solution vector
        """

        #assert self.past_timestep_count != 0, '! pygdh.problem.solve_time_dependent_system: self.past_timestep_count == 0'

        assert ((maximum_time_interval != None and timestep_size != None) or (maximum_time_interval != None and timestep_count != None) or (timestep_count != None and timestep_size != None)), '! Problem.solve_time_dependent_system: Not enough information provided to perform time-dependent simulation.'

        for prblm in self.problem:
        
            # The 0th element of the 'field' member is assumed to contain the 
            # solution at the previous timestep on all 'Grid' objects.
            prblm.initialize_time_dependent_solver()

            # Make the timestep size available to the other methods
            if timestep_size == None:
                timestep_size = maximum_time_interval / float(timestep_count)
            elif timestep_count == None:
                timestep_count = int(round(maximum_time_interval / timestep_size))
            else:
                maximum_time_interval = timestep_size * timestep_count
            
            # Make the timestep size available to the other methods
            prblm.timestep_size = timestep_size / float(self.frequency[prblm])
            prblm.inverse_timestep_size = 1. / timestep_size

            prblm.timestep_index = 0

            prblm.set_initial_conditions()
            
        timestep_index = 0
            
        # Create object to handle all output
        output = pygdh_io.Output(self.grid,output_filename_root,output_types)

        # Now that the domain is fully specified, write out mappings needed to 
        # reconstruct the solution over the domain from the arrays of solution
        # values
        output.write_mappings()
        
        start_time = self.time[0]
        
        # Store initial data and just take the time from the first 'Grid'
        # object
        output.write_fields(0,start_time)

        # # Before rotating the initial conditions into the past solution
        # # position, use them to obtain the initial solution guess
        # if self.input == None or use_predictor == False:
        #     for prblm in self.problem:
        #         prblm.initial_guess = prblm.full_DOF_vector_from_fields(0)

        #     #self.time[0] = 0.0
        # else:
        #     for prblm in self.problem:            
        #         prblm.initial_guess = prblm.initial_guess(self.timestep_size,prblm.full_DOF_vector_from_DOF_vectors(-1))

        print '. Initialization complete, starting time-dependent solver.'

        wall_time_interval = []
        while (self.time[0] - start_time < maximum_time_interval) and (not adaptive_timestepping and timestep_index < timestep_count):

            reference_time = time.time()
            
            # Rotate the initial solution into the -1 position.
            # self.time[-1] is initialized to zero by default            
            self.time.rotate(-1)
            self.oldest_time = self.time[0]

            timestep_index += 1
            for prblm in self.problem:
                prblm.timestep_index += 1
            
            current_time = self.time[-1] + timestep_size
            # Just take the time from the first ``grid.Grid`` object
            if not quiet:
                print "Timestep", timestep_index, "at time", current_time, '...',
            
            for grid in self.grid:
                grid.field.rotate(-1)
                # Save the oldest field and add a new "present" field
                grid.oldest_field = grid.field.popleft()
                grid.field.appendleft(grid.candidate_field)
                        
            # When multiple Problems make use of the the same Grids, broadcasting the full DOF vector for a given Problem does not guarantee that the Grid data for the current timestep is current. Therefore, the Grid data for the current timestep is initially populated with a copy from the last timestep
            for grid in self.grid:
                grid.field[0][0:grid.field_count,0:grid.unit_count] = grid.field[-1][0:grid.field_count,0:grid.unit_count]

            for prblm in self.problem:

                success = prblm.time_dependent_system_step(output,handle_exception_in_solver=handle_exception_in_solver,adaptive_timestepping=adaptive_timestepping,minimum_timestep_size=minimum_timestep_size,use_predictor=use_predictor,quiet=quiet)
                if not success:
                    break

                success = not prblm.abort_solver_and_reject_solution()
                if not success:
                    break

            if not success:
                print '! Stopping time-dependent solver and rejecting most recent solution.'

                ### I don't think that this makes sense, the restored information is never actually used.
                self.time[0] = self.oldest_time
                for grid_obj in self.grid:
                    # Remove the bad solution, this is the same as candidate_field
                    grid_obj.field.popleft()

                    # Restore the oldest solution
                    
                    #self.grid_DOF_vector[grid_obj.number].popleft()
                    grid_obj.field.appendleft(grid_obj.oldest_field)
                    
                    # Rotate the previous solution into the current position
                    grid_obj.field.rotate(1)

                    #self.grid_DOF_vector[grid_obj.number].appendleft(self.grid_oldest_DOF_vector[grid_obj.number])

                    #self.grid_DOF_vector[grid_obj.number].rotate(1)
                    
                output.close()
                    
                return wall_time_interval               
            
            # # This might not be the right place for this
            # for grid_obj in self.grid:
            #     grid_obj.set_boundary_values()

            # The solution is acceptable, so keep the new element and switch the candidate_field members with oldest_field members in preparation for a following timestep
            self.time[0] = self.time[-1] + timestep_size
            for grid_obj in self.grid:
                grid_obj.candidate_field = grid_obj.oldest_field
                
            # This is the user's opportunity to perform validation, and store results besides the solution
            # Maybe this should happen after each stage
            for prblm in self.problem:
                prblm.process_solution()
                    
            # Store solution data, just take the time from the first ``grid.Grid`` object
            output.write_fields(timestep_index,self.time[0])

            for prblm in self.problem:
                if prblm.abort_solver_and_accept_solution():
                    print '! Stopping time-dependent solver and accepting most recent solution.'
                    output.close()
                    return wall_time_interval

        output.close()

        print '. Time-dependent solver method finished.'
        return wall_time_interval
            
